require 'fog/aws'
require 'carrierwave/storage/fog'

#
# This concern should add object storage support
# to the GitlabUploader class
#
module ObjectStorage
  RemoteStoreError = Class.new(StandardError)
  UnknownStoreError = Class.new(StandardError)
  ObjectStoreUnavailable = Class.new(StandardError)

  module Store
    LOCAL = 1
    REMOTE = 2
  end

  module Extension
    # this extension is the glue between the ObjectStorage::Concern and RecordsUploads::Concern
    module RecordsUploads
      extend ActiveSupport::Concern

      included do |base|
        raise ObjectStoreUnavailable, "#{base} must include ObjectStorage::Concern to use extensions."  unless base < Concern
        base.include(::RecordsUploads::Concern)
      end

      def upload=(upload)
        return unless upload

        self.object_store = upload.store
        super
      end
    end
  end

  module Concern
    extend ActiveSupport::Concern

    included do |base|
      base.include(ObjectStorage)

      before :store, :verify_license!
    end

    attr_reader :object_store

    def initialize(model=nil, mounted_as=nil)
      super

      self.upload = model&.try(:"#{mounted_as}_upload", self)
    end

    class_methods do
      def object_store_options
        storage_options&.object_store
      end

      def object_store_enabled?
        object_store_options&.enabled
      end

      def background_upload_enabled?
        object_store_options&.background_upload
      end

      def object_store_credentials
        object_store_options&.connection&.to_hash&.deep_symbolize_keys
      end

      def remote_store_path
        object_store_options&.remote_directory
      end

      def licensed?
        License.feature_available?(:object_storage)
      end
    end

    def file_storage?
      storage.is_a?(CarrierWave::Storage::File)
    end

    def file_cache_storage?
      cache_storage.is_a?(CarrierWave::Storage::File)
    end

    def object_store
      @object_store ||= model.try(store_serialization_column) || Store::LOCAL
    end

    def object_store=(value)
      @object_store = value || Store::LOCAL
      @storage = storage_for(@object_store)
    end

    # Return true if the current file is part or the model (i.e. is mounted in the model)
    #
    def persist_object_store?
      model.respond_to?(:"#{store_serialization_column}=")
    end

    # Save the current @object_store to the model <mounted_as>_store column
    def persist_object_store!
      return unless persist_object_store?

      updated = model.update_column(store_serialization_column, @object_store)
      raise ActiveRecordError unless updated
    end

    def use_file
      if file_storage?
        return yield path
      end

      begin
        cache_stored_file!
        yield cache_path
      ensure
        cache_storage.delete_dir!(cache_path(nil))
      end
    end

    def filename
      super || file&.filename
    end

    #
    # Move the file to another store
    #
    #   new_store: Enum (Store::LOCAL, Store::REMOTE)
    #
    def migrate!(new_store)
      return unless object_store != new_store
      return unless file

      file_to_delete = file
      self.object_store = new_store # this changes the storage and file

      cache_stored_file! if file_storage?

      with_callbacks(:store, file_to_delete) do # for #store_versions!
        storage.store!(file).tap do |new_file|
          @file = new_file
          begin
            # Triggering a model.save! will cause the new_file to be deleted.
            # I still need to investigate exactly why, but this seems like a weird interaction
            # between activerecord and carrierwave
            persist_object_store!
            file_to_delete.delete if new_file.exists?
          rescue => e
            # since we change storage store the new storage
            # in case of failure delete new file
            new_file.delete
            raise e
          end
        end
      end

      file
    end

    def schedule_migration_to_object_storage(*args)
      return unless self.class.object_store_enabled?
      return unless self.class.background_upload_enabled?
      return unless self.class.licensed?
      return unless self.file_storage?

      ObjectStorageUploadWorker.perform_async(self.class.name, model.class.name, mounted_as, model.id)
    end

    def fog_directory
      self.class.remote_store_path
    end

    def fog_credentials
      self.class.object_store_credentials
    end

    def fog_public
      false
    end

    def move_to_store
      return true if Store::LOCAL

      file.try(:storage) == storage
    end

    def move_to_cache
      return true if object_store == Store::LOCAL

      file.try(:storage) == cache_storage
    end

    def verify_license!(_file)
      return if file_storage?

      raise 'Object Storage feature is missing' unless self.class.licensed?
    end

    def exists?
      file.present?
    end

    def cache_dir
      File.join(root, base_dir, 'tmp/cache')
    end

    # Override this if you don't want to save local files by default to the Rails.root directory
    def work_dir
      # Default path set by CarrierWave:
      # https://github.com/carrierwaveuploader/carrierwave/blob/v1.1.0/lib/carrierwave/uploader/cache.rb#L182
      # CarrierWave.tmp_path
      File.join(root, base_dir, 'tmp/work')
    end

    def store_dir(store = nil)
      store_dirs[store || object_store]
    end

    def store_dirs
      {
        Store::LOCAL => File.join(base_dir, dynamic_segment),
        Store::REMOTE => File.join(dynamic_segment)
      }
    end

    private

    def serialization_column
      model.class.uploader_options.dig(mounted_as, :mount_on) || mounted_as
    end

    # Returns the column where the 'store' is saved
    #   defaults to 'store'
    def store_serialization_column
      [serialization_column, 'store'].compact.join('_').to_sym
    end

    def storage
      @storage ||= storage_for(object_store)
    end

    def storage_for(store)
      case store
      when Store::REMOTE
        raise 'Object Storage is not enabled' unless self.class.object_store_enabled?

        CarrierWave::Storage::Fog.new(self)
      when Store::LOCAL
        CarrierWave::Storage::File.new(self)
      else
        raise UnknownStoreError
      end
    end

    # To prevent files in local storage from moving across filesystems, override
    # the default implementation:
    # http://github.com/carrierwaveuploader/carrierwave/blob/v1.1.0/lib/carrierwave/uploader/cache.rb#L181-L183
    def workfile_path(for_file = original_filename)
      # To be safe, keep this directory outside of the the cache directory
      # because calling CarrierWave.clean_cache_files! will remove any files in
      # the cache directory.
      File.join(work_dir, @cache_id, version_name.to_s, for_file)
    end
  end
end
